/*
 * GIMP Speedline Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2013-2015 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "speedline_preview.h"
#include "speedline_parallel.h"
#include "speedline_line.h"

typedef enum _ParallelGuideObj {
	PARALLEL_GUIDE_OBJ_NONE = 0,
	PARALLEL_GUIDE_OBJ_CENTER,
	PARALLEL_GUIDE_OBJ_IN_POINT,
	PARALLEL_GUIDE_OBJ_OUT_POINT,
	PARALLEL_GUIDE_OBJ_IN_LINE,
	PARALLEL_GUIDE_OBJ_OUT_LINE,
	PARALLEL_GUIDE_OBJ_IN_LENGTH_POINT,
	PARALLEL_GUIDE_OBJ_OUT_LENGTH_POINT,
	PARALLEL_GUIDE_OBJ_CONNECTED_LENGTH_LINE,
} ParallelGuideObj;

typedef struct _Point {
	gdouble x;
	gdouble y;
} Point;

#define GET_DISTANCE(X1,Y1,X2,Y2) (sqrt(pow((X2)-(X1),2)+pow((Y2)-(Y1),2)))
#define SWAP(TYPE,A,B) {TYPE temp = (B); (B) = (A); (A) = temp; }

static void path_guide_center (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble out_x, gdouble out_y);
static void path_guide_in_point (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_out_point (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_in_line (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_out_line (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_in_length_point (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_out_length_point (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_connected_length_line (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range);
static void path_guide_in_freehand (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range, KeyValue *points, gint *count);
static void path_guide_out_freehand (cairo_t *cr, gdouble in_x, gdouble in_y, gdouble in_range, gdouble out_x, gdouble out_y, gdouble out_range, KeyValue *points, gint *count);
static ParallelGuideObj path_guide_hit_test (cairo_t *cr, GdkWindow *window, PluginVals	*vals, gdouble x, gdouble y);
static void get_global_vals (PluginVals *vals, gdouble	*in_x, gdouble	*in_y, gdouble	*in_range, gdouble	*out_x, gdouble	*out_y, gdouble	*out_range);


// (1) line ..... y = ax + b
// (2) circle ... (x-c)^2 + (y-d)^2 = D^2
//     line must throught circle center.
static void
get_cross_points_line_circle_center (gdouble a,
									 gdouble b,
									 gdouble c,
									 gdouble d,
									 gdouble D,
									 Point *point1,
									 Point *point2)
{
	point1->x = D/sqrt(1 + a*a) + c;
	point1->y = a * D/sqrt(1 + a*a) + d;
	point2->x = -D/sqrt(1 + a*a) + c;
	point2->y = a * -D/sqrt(1 + a*a) + d;
}

static void
get_2points (Point		*in_st,
			 Point		*in_ed,
			 Point		*out_st,
			 Point		*out_ed,
			 gdouble	percent,
			 Point		*st,
			 Point		*ed)
{
	st->x = (in_ed->x - in_st->x) * percent + in_st->x;
	st->y = (in_ed->y - in_st->y) * percent + in_st->y;
	ed->x = (out_ed->x - out_st->x) * percent + out_st->x;
	ed->y = (out_ed->y - out_st->y) * percent + out_st->y;
}

static void
get_4points (gdouble in_x,
			 gdouble in_y,
			 gdouble out_x,
			 gdouble out_y,
			 gdouble in_range,
			 gdouble out_range,
			 Point		*in_st,
			 Point		*in_ed,
			 Point		*out_st,
			 Point		*out_ed)
{	
	if (in_x == out_x) {
		if (in_y < out_y) {
			in_st->x = in_x - in_range / 2;
			in_st->y = in_y;
			in_ed->x = in_x + in_range / 2;
			in_ed->y = in_y;
			out_st->x = out_x - out_range;
			out_st->y = out_y;
			out_ed->x = out_x + out_range;
			out_ed->y = out_y;
		} else {
			in_st->x = in_x + in_range / 2;
			in_st->y = in_y;
			in_ed->x = in_x - in_range / 2;
			in_ed->y = in_y;
			out_st->x = out_x + out_range / 2;
			out_st->y = out_y;
			out_ed->x = out_x - out_range / 2;
			out_ed->y = out_y;
		}
	} else if (in_y == out_y) {
		if (in_x > out_x) {
			in_st->x = in_x;
			in_st->y = in_y - in_range / 2;
			in_ed->x = in_x;
			in_ed->y = in_y + in_range / 2;
			out_st->x = out_x;
			out_st->y = out_y - out_range / 2;
			out_ed->x = out_x;
			out_ed->y = out_y + out_range / 2;
		} else {
			in_st->x = in_x;
			in_st->y = in_y + in_range / 2;
			in_ed->x = in_x;
			in_ed->y = in_y - in_range / 2;
			out_st->x = out_x;
			out_st->y = out_y + out_range / 2;
			out_ed->x = out_x;
			out_ed->y = out_y - out_range / 2;
		}
	} else {
#if 1
		gdouble a, b, c, d;

		a = -1 / ((out_y - in_y) / (out_x - in_x));
		b = in_y - a * in_x;
		c = in_x;
		d = in_y;
		get_cross_points_line_circle_center (a, b, c, d, in_range/2, in_st, in_ed);

		a = -1 / ((out_y - in_y) / (out_x - in_x));
		b = out_y - a * out_x;
		c = out_x;
		d = out_y;
		get_cross_points_line_circle_center (a, b, c, d, out_range/2, out_st, out_ed);
		{
			gdouble rad = atan2(out_y - in_y, out_x - in_x);
			if (rad >= 0 && rad <= G_PI) {
				SWAP (Point, *in_st, *in_ed);
				SWAP (Point, *out_st, *out_ed);
			}
		}
#else
		//FIXME:
		gdouble rad;
		rad = atan2(out_y - in_y, out_x - in_x);
		in_st->x = cos(rad + G_PI_2) * in_range / 2 + in_x;
		in_st->y = sin(rad + G_PI_2) * in_range / 2 + in_y;
		in_ed->x = cos(rad - G_PI_2) * in_range / 2 + in_x;
		in_ed->y = sin(rad - G_PI_2) * in_range / 2 + in_y;

		out_st->x = cos(rad + G_PI_2) * out_range / 2 + out_x;
		out_st->y = sin(rad + G_PI_2) * out_range / 2 + out_y;
		out_ed->x = cos(rad - G_PI_2) * out_range / 2 + out_x;
		out_ed->y = sin(rad - G_PI_2) * out_range / 2 + out_y;
#endif
	}
}

static void
path_guide_center (cairo_t *cr,
				   gdouble in_x,
				   gdouble in_y,
				   gdouble out_x,
				   gdouble out_y)
{
	gdouble size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;
	cairo_arc (cr, (in_x + out_x) / 2, (in_y + out_y) / 2, preview_local_to_global (size), 0, 2 * G_PI);
}

static void
path_guide_in_point (cairo_t *cr,
					 gdouble in_x,
					 gdouble in_y,
					 gdouble in_range,
					 gdouble out_x,
					 gdouble out_y,
					 gdouble out_range)
{
	gdouble size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;
	cairo_arc (cr, in_x, in_y, preview_local_to_global (size), 0, 2 * G_PI);
}

static void
path_guide_out_point (cairo_t *cr,
					  gdouble in_x,
					  gdouble in_y,
					  gdouble in_range,
					  gdouble out_x,
					  gdouble out_y,
					  gdouble out_range)
{
	gdouble size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;
	cairo_arc (cr, out_x, out_y, preview_local_to_global (size), 0, 2 * G_PI);
}

static void
path_guide_in_line (cairo_t *cr,
					gdouble in_x,
					gdouble in_y,
					gdouble in_range,
					gdouble out_x,
					gdouble out_y,
					gdouble out_range)
{
	cairo_save (cr);
	cairo_translate (cr, in_x, in_y);
	cairo_rotate (cr, G_PI_2 + atan2 (out_y - in_y, out_x - in_x));
	cairo_move_to (cr, in_range / -2.0, 0);
	cairo_line_to (cr, in_range /  2.0, 0);
	cairo_restore (cr);
}

static void
path_guide_out_line (cairo_t *cr,
					 gdouble in_x,
					 gdouble in_y,
					 gdouble in_range,
					 gdouble out_x,
					 gdouble out_y,
					 gdouble out_range)
{
	cairo_save (cr);
	cairo_translate (cr, out_x, out_y);
	cairo_rotate (cr, G_PI_2 + atan2 (in_y - out_y, in_x - out_x));
	cairo_move_to (cr, out_range / -2.0, 0);
	cairo_line_to (cr, out_range /  2.0, 0);
	cairo_restore (cr);
}

static void
path_guide_in_length_point (cairo_t *cr,
							gdouble in_x,
							gdouble in_y,
							gdouble in_range,
							gdouble out_x,
							gdouble out_y,
							gdouble out_range)
{
	gdouble size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;

	cairo_save (cr);
	cairo_translate (cr, in_x, in_y);
	cairo_rotate (cr, G_PI_2 + atan2 (out_y - in_y, out_x - in_x));
	cairo_arc (cr, in_range / -2.0, 0, preview_local_to_global (size), 0, 2 * G_PI);
	cairo_arc (cr, in_range /  2.0, 0, preview_local_to_global (size), 0, 2 * G_PI);
	cairo_restore (cr);
}

static void
path_guide_out_length_point (cairo_t *cr,
							 gdouble in_x,
							 gdouble in_y,
							 gdouble in_range,
							 gdouble out_x,
							 gdouble out_y,
							 gdouble out_range)
{
	gdouble size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;

	cairo_save (cr);
	cairo_translate (cr, out_x, out_y);
	cairo_rotate (cr, G_PI_2 + atan2 (in_y - out_y, in_x - out_x));
	cairo_arc (cr, out_range / -2.0, 0, preview_local_to_global (size), 0, 2 * G_PI);
	cairo_arc (cr, out_range /  2.0, 0, preview_local_to_global (size), 0, 2 * G_PI);
	cairo_restore (cr);
}

static void
path_guide_connected_length_line (cairo_t *cr,
								  gdouble in_x,
								  gdouble in_y,
								  gdouble in_range,
								  gdouble out_x,
								  gdouble out_y,
								  gdouble out_range)
{
	gdouble len;

	len = GET_DISTANCE (in_x, in_y, out_x, out_y);

	cairo_save (cr);
	cairo_translate (cr, in_x, in_y);
	cairo_rotate (cr, G_PI_2 + atan2 (out_y - in_y, out_x - in_x));
	cairo_move_to (cr,  in_range / -2.0, 0);
	cairo_line_to (cr, out_range / -2.0, -len);
	cairo_move_to (cr,  in_range /  2.0, 0);
	cairo_line_to (cr, out_range /  2.0, -len);
	cairo_restore (cr);
}

static void
path_guide_in_freehand (cairo_t *cr,
						gdouble in_x,
						gdouble in_y,
						gdouble in_range,
						gdouble out_x,
						gdouble out_y,
						gdouble out_range,
						KeyValue *points,
						gint *count)
{
	gint i;
	gdouble x, y;
	Point inst, ined, outst, outed;
	gdouble percent1 = 1 / FREEHAND_POINTS_MAX;

	if (*count == 0) {
		parallel_reset_freehand (points, count);
	}

	get_4points (in_x, in_y, out_x, out_y, in_range, out_range, &inst, &ined, &outst, &outed);

	cairo_save (cr);
	for (i = 0; i < *count; i++) {
		gdouble percent = points[i].key;
		gdouble rate = points[i].value;
		Point p1, p2;

		get_2points (&inst, &ined, &outst, &outed, percent, &p1, &p2);
		x = (p2.x - p1.x) * (1 - rate) + p1.x;
		y = (p2.y - p1.y) * (1 - rate) + p1.y;

		if (i == 0) {
			cairo_move_to (cr, x, y);
		} else {
			cairo_line_to (cr, x, y);
		}
	}
	cairo_restore (cr);
}

static void
path_guide_out_freehand (cairo_t *cr,
						 gdouble in_x,
						 gdouble in_y,
						 gdouble in_range,
						 gdouble out_x,
						 gdouble out_y,
						 gdouble out_range,
						 KeyValue *points,
						 gint *count)
{
	gint i;
	gdouble x, y;
	Point inst, ined, outst, outed;
	gdouble percent1 = 1 / FREEHAND_POINTS_MAX;

	if (*count == 0) {
		parallel_reset_freehand (points, count);
	}

	get_4points (in_x, in_y, out_x, out_y, in_range, out_range, &inst, &ined, &outst, &outed);

	cairo_save (cr);
	for (i = 0; i < *count; i++) {
		gdouble percent = points[i].key;
		gdouble rate = points[i].value;
		Point p1, p2;

		get_2points (&inst, &ined, &outst, &outed, percent, &p1, &p2);
		x = (p2.x - p1.x) * rate + p1.x;
		y = (p2.y - p1.y) * rate + p1.y;

		if (i == 0) {
			cairo_move_to (cr, x, y);
		} else {
			cairo_line_to (cr, x, y);
		}
	}
	cairo_restore (cr);
}

static ParallelGuideObj
path_guide_hit_test (cairo_t	*cr,
					 GdkWindow	*window,
					 PluginVals	*vals,
					 gdouble	x,
					 gdouble	y)
{
	gdouble in_x, in_y, in_range, out_x, out_y, out_range;

	get_global_vals (vals, &in_x, &in_y, &in_range, &out_x, &out_y, &out_range);

	cairo_set_source_rgba (cr, 1, 1, 1, 1);
	cairo_set_line_width (cr, preview_local_to_global (PREVIEW_GUIDE_LINE_WIDTH_FOR_HITTEST));
	cairo_device_to_user (cr, &x, &y);

	//check if center
	path_guide_center (cr, in_x, in_y, out_x, out_y);
	if (cairo_in_fill (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_CENTER;
	}
	cairo_new_path (cr);

	//check if in_point
	path_guide_in_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_fill (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_IN_POINT;
	}
	cairo_new_path (cr);

	//check if out_point
	path_guide_out_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_fill (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_OUT_POINT;
	}
	cairo_new_path (cr);
	
	//check if in_line
	path_guide_in_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_stroke (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_IN_LINE;
	}
	cairo_new_path (cr);

	//check if out_line
	path_guide_out_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_stroke (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_OUT_LINE;
	}
	cairo_new_path (cr);

	//check if in_length_point
	path_guide_in_length_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_fill (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_IN_LENGTH_POINT;
	}
	cairo_new_path (cr);

	//check if out_length_point
	path_guide_out_length_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_fill (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_OUT_LENGTH_POINT;
	}
	cairo_new_path (cr);

	//check if connected length line
	path_guide_connected_length_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	if (cairo_in_stroke (cr, x, y)) {
		return PARALLEL_GUIDE_OBJ_CONNECTED_LENGTH_LINE;
	}
	cairo_new_path (cr);

	return PARALLEL_GUIDE_OBJ_NONE;
}

static void 
get_global_vals (PluginVals *vals,
				 gdouble	*in_x,
				 gdouble	*in_y,
				 gdouble	*in_range,
				 gdouble	*out_x,
				 gdouble	*out_y,
				 gdouble	*out_range)
{
	gdouble resx, resy;
	gimp_image_get_resolution (vals->image_ID, &resx, &resy);

	*in_x = gimp_units_to_pixels (vals->parallel_in_x, vals->area_unit, resx);
	*in_y = gimp_units_to_pixels (vals->parallel_in_y, vals->area_unit, resx);
	*in_range = gimp_units_to_pixels (vals->parallel_in_range, vals->area_unit, resx);
	*out_x = gimp_units_to_pixels (vals->parallel_out_x, vals->area_unit, resx);
	*out_y = gimp_units_to_pixels (vals->parallel_out_y, vals->area_unit, resx);
	*out_range = gimp_units_to_pixels (vals->parallel_out_range, vals->area_unit, resx);
}

static gdouble
get_freehand_percent (gdouble in_x,
					  gdouble in_y,
					  gdouble in_range,
					  gdouble out_x,
					  gdouble out_y,
					  gdouble out_range,
					  gdouble x,
					  gdouble y) {
	gdouble Drange;
	Point D, Dst, Ded;

	// D(Dx,Dy)
	{
		if (out_x == in_x) {
			D.x = in_x;
			D.y = y;
		} else if (out_y == in_y) {
			D.x = x;
			D.y = in_y;
		} else {
			// y = ax + b ... in -> out
			// y = cx + d ... (x, y) -> (Dx, Dy)
			gdouble a, b, c, d;
			a = (out_y - in_y) / (out_x - in_x);
			b = in_y - a * in_x;
			c = -1 / a;
			d = y - c * x;
			D.x = (d -b) / (a - c);
			D.y = c * D.x + d;
			d = y - c * x;
			D.x = (d -b) / (a - c);
			D.y = c * D.x + d;
		}
	}
	
	// Drange, Dst(Dstx, Dsty), Ded(Dedx, Dedy)
	{
		gdouble d1, d2;
		d1 = GET_DISTANCE (out_x, out_y, D.x, D.y);
		d2 = GET_DISTANCE (out_x, out_y, in_x, in_y);
		if (GET_DISTANCE (in_x, in_y, D.x, D.y) > d2) {
			d1 = -d1;
		}
		Drange = (d1 / d2) * (in_range - out_range) + out_range;

		if (out_x == in_x) {
			Dst.x = D.x + Drange / 2;
			Dst.y = y;
			Ded.x = D.x - Drange / 2;
			Ded.y = y;
		} else if (out_y == in_y) {
			Dst.x = x;
			Dst.y = D.y + Drange / 2;
			Ded.x = x;
			Ded.y = D.y - Drange / 2;
		} else {
			gdouble a, b, c, d;
			a = -1 / ((out_y - in_y) / (out_x - in_x));
			b = D.y - a * D.x;
			c = D.x;
			d = D.y;
			get_cross_points_line_circle_center (a, b, c, d, Drange/2, &Dst, &Ded);
		}
	}

	// Get percent
	{
		gdouble rad;
		gdouble d1, d2, percent;


		d1 = GET_DISTANCE (Dst.x, Dst.y, x, y);
		d2 = GET_DISTANCE (Ded.x, Ded.y, x, y);
		percent = d1 / Drange;

		if (d1 > Drange || d2 > Drange) {
			return -1;
		}

		rad = atan2 (out_y - in_y, out_x - in_x);
		if (rad > 0 && rad <= G_PI) {
			percent = 1 - percent;
		}
		return percent;
	}
}

static void
update_freehand_points (PluginVals				*vals,
						FreehandEditModeType	type,
						gdouble					stx,
						gdouble					sty,
						gdouble					edx,
						gdouble					edy)
{
	gint i;
	gdouble in_x = vals->parallel_in_x;
	gdouble in_y = vals->parallel_in_y;
	gdouble in_range = vals->parallel_in_range;
	gdouble out_x = vals->parallel_out_x;
	gdouble out_y = vals->parallel_out_y;
	gdouble out_range = vals->parallel_out_range;
	KeyValue *points;
	gint points_count;
	Point inst, ined, outst, outed;
	gdouble percent_st, percent_ed, percent1, rate_st, rate_ed;

	// fill basic parameters
	switch (type) {
		case FREEHAND_EDIT_MODE_PARALLEL_IN:
			points = vals->parallel_in_freehand_points;
			points_count = vals->parallel_in_freehand_points_count;
			break;
		case FREEHAND_EDIT_MODE_PARALLEL_OUT:
			points = vals->parallel_out_freehand_points;
			points_count = vals->parallel_out_freehand_points_count;
			break;
		default:
			return;
	}

	get_4points (in_x, in_y, out_x, out_y, in_range, out_range, &inst, &ined, &outst, &outed);

	// Get percent
	percent_st = get_freehand_percent (in_x, in_y, in_range, out_x, out_y, out_range, stx, sty);
	percent_ed = get_freehand_percent (in_x, in_y, in_range, out_x, out_y, out_range, edx, edy);
	if (percent_st < 0 || percent_ed < 0) return;

	// Get rate
	{
		Point p1_st, p2_st, p1_ed, p2_ed;
		gdouble dist_all_st, dist_all_ed;

		get_2points (&inst, &ined, &outst, &outed, percent_st, &p1_st, &p2_st);
		get_2points (&inst, &ined, &outst, &outed, percent_ed, &p1_ed, &p2_ed);
		
		dist_all_st = GET_DISTANCE (p1_st.x, p1_st.y, p2_st.x, p2_st.y);
		dist_all_ed = GET_DISTANCE (p1_ed.x, p1_ed.y, p2_ed.x, p2_ed.y);

		switch (type) {
			case FREEHAND_EDIT_MODE_PARALLEL_IN:
				rate_st = GET_DISTANCE (p2_st.x, p2_st.y, stx, sty) / dist_all_st;
				rate_ed = GET_DISTANCE (p2_ed.x, p2_ed.y, edx, edy) / dist_all_ed;
				break;

			case FREEHAND_EDIT_MODE_PARALLEL_OUT:
				rate_st = GET_DISTANCE (p1_st.x, p1_st.y, stx, sty) / dist_all_st;
				rate_ed = GET_DISTANCE (p1_ed.x, p1_ed.y, edx, edy) / dist_all_ed;
				break;
		}
	}
	
	percent1 = 1.0 / points_count;
	if (percent_ed < percent_st) {
			SWAP (gdouble, percent_st, percent_ed);
			SWAP (gdouble, rate_st, rate_ed);
	}

	if (percent_ed - percent_st >= percent1) {
		gdouble n = (rate_ed - rate_st) / ((percent_st - percent_ed) / percent1);
		gdouble nn = 0;

		for (i = 0; i < points_count; i++) {
			gdouble percent = points[i].key;
			if (percent >= percent_st && percent <= percent_ed) {
				points[i].value = rate_st - nn;
				nn += n;
			}
		}
	} else {
		for (i = 0; i < points_count; i++) {
			gdouble percent = points[i].key;
			if (percent >= percent_st) {
				points[i].value = rate_st;
				break;
			}
		}
	}
}

static gdouble
get_freehand_rad_value (gdouble		percent, 
						KeyValue	*points, 
						gint		count)
{
	gint i;
	gdouble percent1 = G_PI * 2 / count;

	if (percent < 0) { percent = 0; }
	if (percent > 1) { percent = 1; }

	for (i = 0; i < count; i++) {
		if (points[i].key >= percent) {
			return points[i].value;
		}
	}
	return 0;
}

void
parallel_draw_guide (cairo_t	*cr,
					 PluginVals	*vals)
{
	gdouble in_x, in_y, in_range, out_x, out_y, out_range;

	get_global_vals (vals, &in_x, &in_y, &in_range, &out_x, &out_y, &out_range);

	cairo_set_line_width (cr, preview_local_to_global (PREVIEW_GUIDE_LINE_WIDTH));
	cairo_set_source_rgba (cr,
		PREVIEW_GUIDE_COLOR_R,
		PREVIEW_GUIDE_COLOR_G,
		PREVIEW_GUIDE_COLOR_B,
		1);

	// in freehand
	if (vals->parallel_in_freehand_active) {
	path_guide_in_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
		path_guide_in_freehand (cr, in_x, in_y, in_range, out_x, out_y, out_range,
								vals->parallel_in_freehand_points,
								&vals->parallel_in_freehand_points_count);
		cairo_save (cr);
		{
			gdouble dashes[] = {6, 3};
			if (preview_freehand_edit_mode_get () == FREEHAND_EDIT_MODE_PARALLEL_IN) {
				cairo_set_source_rgba (cr,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_R,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_G,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_B,
					1);
			}
			cairo_set_dash (cr, dashes, 2, 0.0);
			cairo_stroke (cr);
		}
		cairo_restore (cr);
	}

	// out freehand
	if (vals->parallel_out_freehand_active) {
		path_guide_out_freehand (cr, in_x, in_y, in_range, out_x, out_y, out_range,
								vals->parallel_out_freehand_points,
								&vals->parallel_out_freehand_points_count);
		cairo_save (cr);
		{
			gdouble dashes[] = {6, 3};
			if (preview_freehand_edit_mode_get () == FREEHAND_EDIT_MODE_PARALLEL_OUT) {
				cairo_set_source_rgba (cr,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_R,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_G,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_B,
					1);
			}
			cairo_set_dash (cr, dashes, 2, 0.0);
			cairo_stroke (cr);
		}
		cairo_restore (cr);
	}

	//center
	path_guide_center (cr, in_x, in_y, out_x, out_y);
	cairo_fill (cr);

	//in point
	path_guide_in_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_fill (cr);

	//out point
	path_guide_out_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_fill (cr);

	//in line
	path_guide_in_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_stroke (cr);

	//out line
	path_guide_out_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_stroke (cr);

	//in length point
	path_guide_in_length_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_fill (cr);

	//out length point
	path_guide_out_length_point (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_fill (cr);

	//connected length line
	path_guide_connected_length_line (cr, in_x, in_y, in_range, out_x, out_y, out_range);
	cairo_stroke (cr);


	//lines
	{
		gint n, nmax;
		gdouble ylen;
		ylen = GET_DISTANCE (in_x, in_y, out_x, out_y);

		cairo_save (cr);
		cairo_translate (cr, in_x, in_y);
		cairo_rotate (cr, G_PI_2 + atan2 (out_y - in_y, out_x - in_x));
		nmax = PREVIEW_GUIDE_LINE_NUMBER;
		for (n = 0; n < nmax; n++) {
			cairo_move_to (cr,  (in_range  / -2.0) + in_range  * ((gdouble)n/nmax), 0);
			cairo_line_to (cr,  (out_range / -2.0) + out_range * ((gdouble)n/nmax), -ylen);
		}
		cairo_restore (cr);
		cairo_stroke (cr);
	}
}

void
parallel_draw_line (cairo_t		*cr,
					PluginVals	*vals)
{
	LineObj *line = line_new (vals);
	gdouble in_x, in_y, in_range, out_x, out_y, out_range;
	gdouble block_sum = 0;
	gdouble wave_len = 0;

	get_global_vals (vals, &in_x, &in_y, &in_range, &out_x, &out_y, &out_range);

	cairo_save (cr);
	{
		gdouble p1, psum;
		gdouble max_len = 0;
		gdouble distance_in_out = 0;

		distance_in_out = GET_DISTANCE (in_x, in_y, out_x, out_y);

		//compute length for line_pitch
		switch (vals->line_pitch_type) {
			case LINE_PITCH_TYPE_IN:  max_len = in_range; break;
			case LINE_PITCH_TYPE_OUT: max_len = out_range; break;
		}
		p1 = line->pitch / max_len;
		psum = 0;

		g_assert (p1 != 0);

		if (line_is_wave_mode (line) == TRUE) {
			 wave_len = in_range / (line_wave_npoint_get (line) - 1);
		}

		//normalize area center and angle
		cairo_translate (cr, in_x, in_y);
		cairo_rotate (cr, atan2 ((out_y - in_y), (out_x - in_x)));
		cairo_rotate (cr, -G_PI_2);

		while (psum <= in_range) {
			gdouble p = p1 * in_range;
			gdouble inx, iny, outx, outy;

			p = p - (p * g_rand_double (line->randobj_pitch) * line->pitch_rand / 100.0);
			inx = -in_range / 2.0 + psum;
			iny = 0;
			outx = -out_range / 2.0 + out_range * (psum / in_range);
			outy = distance_in_out;

			if (line_is_block_mode (line) == TRUE) {
				if (psum >= block_sum) {
					line_block_state_invert(line);
					if (line_block_state_get (line) == TRUE) {
						block_sum += line_block_on_pitch_get (line);
					} else {
						block_sum += line_block_off_pitch_get (line);
					}
				}
				if (line_block_state_get (line) == FALSE) {
					goto block_skip;
				}
			}

			cairo_save (cr);
			{
				//normaze line center and angle
				gdouble line_length;
				gdouble line_length_adj = 0;
				line_length = GET_DISTANCE (inx, iny, outx, outy);

				cairo_translate (cr, inx, iny);
				cairo_rotate (cr, atan2 (outy - iny, outx - inx));

				//Wave mode
				if (line_is_wave_mode (line) == TRUE) {
					gdouble r = fmod(psum, wave_len) / wave_len;
					gdouble depth = line_wave_depth_get (line, r);
					cairo_translate (cr, depth, 0);
				}
				// Freehand mode In
				if (vals->parallel_in_freehand_active) {
					gdouble rate = get_freehand_rad_value (psum / in_range,
														   vals->parallel_in_freehand_points,
														   vals->parallel_in_freehand_points_count);
					gdouble adj = line_length - line_length * rate;
					cairo_translate (cr, adj, 0);
					line_length_adj -= adj;
				}

				// Freehand mode Out
				if (vals->parallel_out_freehand_active) {
					gdouble rate = get_freehand_rad_value (psum / in_range,
														   vals->parallel_out_freehand_points,
														   vals->parallel_out_freehand_points_count);
					gdouble adj = line_length * (1 - rate);
					line_length_adj -= adj;
				}

				line_draw (line, cr, line_length + line_length_adj);
			}
			cairo_restore (cr);

block_skip:

			psum += p;
		}
	}
	cairo_restore (cr);

	line_destroy (line);
}

gboolean
parallel_event_hook (cairo_t	*cr,
					 GdkEvent	*event,
					 PluginVals	*vals)
{
	static ParallelGuideObj dragging_obj = PARALLEL_GUIDE_OBJ_NONE;
	static gdouble dragging_start_point_x = 0;
	static gdouble dragging_start_point_y = 0;
	static gdouble dragging_start_in_x = 0;
	static gdouble dragging_start_in_y = 0;
	static gdouble dragging_start_out_x = 0;
	static gdouble dragging_start_out_y = 0;
	static gdouble dragging_start_in_range = 0;
	static gdouble dragging_start_out_range = 0;
	static gdouble dragging_start_freehand_point_x = 0;
	static gdouble dragging_start_freehand_point_y = 0;

	switch (event->type) {
		case GDK_EXPOSE:
			break;

		case GDK_BUTTON_PRESS:
			dragging_start_point_x = event->button.x;
			dragging_start_point_y = event->button.y;
			cairo_device_to_user (cr, &dragging_start_point_x, &dragging_start_point_y);
			dragging_start_in_x = vals->parallel_in_x;
			dragging_start_in_y = vals->parallel_in_y;
			dragging_start_out_x = vals->parallel_out_x;
			dragging_start_out_y = vals->parallel_out_y;
			dragging_start_in_range = vals->parallel_in_range;
			dragging_start_out_range = vals->parallel_out_range;

			if (preview_freehand_is_editing ()) {
				dragging_start_freehand_point_x = dragging_start_point_x;
				dragging_start_freehand_point_y = dragging_start_point_y;
			} else {
				dragging_obj = path_guide_hit_test (cr,
													event->any.window,
													vals,
													event->button.x,
													event->button.y);
			}
			break;

		case GDK_BUTTON_RELEASE:
			dragging_start_point_x = 0;
			dragging_start_point_y = 0;
			dragging_start_in_x = 0;
			dragging_start_in_y = 0;
			dragging_start_out_x = 0;
			dragging_start_out_y = 0;
			dragging_start_in_range = 0;
			dragging_start_out_range = 0;
			dragging_obj = PARALLEL_GUIDE_OBJ_NONE;
			dragging_start_freehand_point_x = 0;
			dragging_start_freehand_point_y = 0;
			break;

		case GDK_MOTION_NOTIFY:
			{
				gdouble gx, gy;
				gdouble resx, resy;
				gimp_image_get_resolution (vals->image_ID, &resx, &resy);
				gx = event->motion.x;
				gy = event->motion.y;
				cairo_device_to_user (cr, &gx, &gy);

				if (preview_freehand_is_editing ()) {
					update_freehand_points (vals,
											preview_freehand_edit_mode_get (),
											dragging_start_freehand_point_x,
											dragging_start_freehand_point_y,
											gx, gy);
					dragging_start_freehand_point_x = gx;
					dragging_start_freehand_point_y = gy;
				} else {
					switch (dragging_obj) {
						case PARALLEL_GUIDE_OBJ_CENTER:
							{
								gdouble mx, my;
								mx = gimp_pixels_to_units (gx - dragging_start_point_x, vals->area_unit, resx);
								my = gimp_pixels_to_units (gy - dragging_start_point_y, vals->area_unit, resy);

								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_x), mx + dragging_start_in_x);
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_y), my + dragging_start_in_y);
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_x), mx + dragging_start_out_x);
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_y), my + dragging_start_out_y);
							}
							break;

						case PARALLEL_GUIDE_OBJ_IN_POINT:
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_x), gimp_pixels_to_units (gx, vals->area_unit, resx));
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_y), gimp_pixels_to_units (gy, vals->area_unit, resy));
							break;

						case PARALLEL_GUIDE_OBJ_OUT_POINT:
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_x), gimp_pixels_to_units (gx, vals->area_unit, resx));
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_y), gimp_pixels_to_units (gy, vals->area_unit, resy));
							break;

						case PARALLEL_GUIDE_OBJ_IN_LINE:
							{
								gdouble valx, valy;

								if (dragging_start_in_x == dragging_start_out_x) {
									valx = dragging_start_in_x;
									valy = gy;
								} else if (dragging_start_in_y == dragging_start_out_y) {
									valx = gx;
									valy = dragging_start_in_y;
								} else {
									gdouble nx, ny;
									gdouble m1, n1;// y = (m1*x)+n1 ... line1;
									gdouble m2, n2;// y = (m2*x)+n2 ... line2;
									nx = gimp_pixels_to_units (gx, vals->area_unit, resx);
									ny = gimp_pixels_to_units (gy, vals->area_unit, resy);

									m1 = (dragging_start_out_y - dragging_start_in_y) / (dragging_start_out_x - dragging_start_in_x);
									n1 = dragging_start_in_y - (m1 * dragging_start_in_x);
									m2 = -1.0 / m1;
									n2 = ny - m2 * nx;

									//cross point of line1 and line2
									valx = (n2 - n1) / (m1 - m2);
									valy = valx * m1 + n1;
								}
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_x), valx);
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_y), valy);
							}
							break;

						case PARALLEL_GUIDE_OBJ_OUT_LINE:
							{
								gdouble valx, valy;

								if (dragging_start_in_x == dragging_start_out_x) {
									valx = dragging_start_in_x;
									valy = gy;
								} else if (dragging_start_in_y == dragging_start_out_y) {
									valx = gx;
									valy = dragging_start_in_y;
								} else {
									gdouble nx, ny;
									gdouble m1, n1;// y = (m1*x)+n1 ... line1;
									gdouble m2, n2;// y = (m2*x)+n2 ... line2;
									nx = gimp_pixels_to_units (gx, vals->area_unit, resx);
									ny = gimp_pixels_to_units (gy, vals->area_unit, resy);

									m1 = (dragging_start_in_y - dragging_start_out_y) / (dragging_start_in_x - dragging_start_out_x);
									n1 = dragging_start_out_y - (m1 * dragging_start_out_x);
									m2 = -1.0 / m1;
									n2 = ny - m2 * nx;

									//cross point of line1 and line2
									valx = (n2 - n1) / (m1 - m2);
									valy = valx * m1 + n1;
								}
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_x), valx);
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_y), valy);
							}
							break;

						case PARALLEL_GUIDE_OBJ_IN_LENGTH_POINT:
							{
								gdouble px, py, nx, ny, val;

								px = gimp_pixels_to_units (vals->parallel_in_x, vals->area_unit, resx);
								py = gimp_pixels_to_units (vals->parallel_in_y, vals->area_unit, resy);
								nx = gimp_pixels_to_units (gx, vals->area_unit, resx);
								ny = gimp_pixels_to_units (gy, vals->area_unit, resy);
								val = GET_DISTANCE (px, py, nx, ny) * 2;
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_range), val);
							}
							break;

						case PARALLEL_GUIDE_OBJ_OUT_LENGTH_POINT:
							{
								gdouble px, py, nx, ny, val;

								px = gimp_pixels_to_units (vals->parallel_out_x, vals->area_unit, resx);
								py = gimp_pixels_to_units (vals->parallel_out_y, vals->area_unit, resy);
								nx = gimp_pixels_to_units (gx, vals->area_unit, resx);
								ny = gimp_pixels_to_units (gy, vals->area_unit, resy);
								val = GET_DISTANCE (px, py, nx, ny) * 2;
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_range), val);
							}
							break;

						case PARALLEL_GUIDE_OBJ_CONNECTED_LENGTH_LINE:
							{
								gdouble crx1, cry1, crx2, cry2;
								gdouble len, len_in, len_out;
								gdouble nsx, nsy, nex, ney;

								nsx = dragging_start_point_x;
								nsy = dragging_start_point_y;

								nsx = gimp_pixels_to_units (nsx, vals->area_unit, resx);
								nsy = gimp_pixels_to_units (nsy, vals->area_unit, resy);
								nex = gimp_pixels_to_units (gx, vals->area_unit, resx);
								ney = gimp_pixels_to_units (gy, vals->area_unit, resy);

								if (dragging_start_in_x == dragging_start_out_x) {
									crx1 = dragging_start_in_x;
									cry1 = nsy;
									crx2 = nex;
									cry2 = nsy;
								} else if (dragging_start_in_y == dragging_start_out_y) {
									crx1 = nsx;
									cry1 = dragging_start_in_y;
									crx2 = nsx;
									cry2 = ney;
								} else {
									gdouble m1, n1;// y = (m1*x)+n1 ... line1;
									gdouble m2, n2;// y = (m2*x)+n2 ... line2;
									gdouble m3, n3;// y = (m3*x)+n3 ... line3;

									m1 = (dragging_start_in_y - dragging_start_out_y) / (dragging_start_in_x - dragging_start_out_x);
									n1 = dragging_start_out_y - (m1 * dragging_start_out_x);
									m2 = -1.0 / m1;
									n2 = nsy - m2 * nsx;
									m3 = m1;
									n3 = ney - m3 * nex;

									//cross point of line1 and line2
									crx1 = (n2 - n1) / (m1 - m2);
									cry1 = crx1 * m1 + n1;

									//cross point of line2 and line3
									crx2 = (n3 - n2) / (m2 - m3);
									cry2 = crx2 * m2 + n2;
								}

								len = GET_DISTANCE (crx1, cry1, crx2, cry2) - GET_DISTANCE (crx1, cry1, nsx, nsy);
								len *= 2;

								if (dragging_start_in_range + len > 1 &&
									dragging_start_out_range + len > 1) {
									len_in = dragging_start_in_range + len;
									len_out = dragging_start_out_range + len;
									gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_in_range), len_in);
									gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.parallel_out_range), len_out);
								}
							}
							break;
					}
				}
				gtk_widget_queue_draw (widgets.preview);
			}
			break;

		case GDK_KEY_PRESS:
			break;

		case GDK_KEY_RELEASE:
			break;

		default:
			break;
	}
	return FALSE;
}

void
parallel_reset_freehand (KeyValue	*points,
						 gint		*count)
{
	gdouble p1 = 1.0 / FREEHAND_POINTS_MAX;
	gint i = 0;
	*count = FREEHAND_POINTS_MAX;

	for (i = 0; i < FREEHAND_POINTS_MAX; i++) {
		points[i].key = i * p1;
		points[i].value = 1.0;
	}
}
/* vim: set fdm=marker ts=4 sw=4 noet :*/
