/*
 * GIMP Speedline Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2013-2015 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "speedline_preview.h"

#include "speedline_radiation.h"
#include "speedline_parallel.h"

static const gint			g_preview_width  = 500;
static const gint			g_preview_height = 500;

static GtkWidget			*g_preview = NULL;
static GdkPixbuf			*g_preview_buf = NULL;
static gint					g_preview_zoom = 100;
static GdkRectangle			g_preview_orig_buf = {0, 0, 0, 0};
static GdkRectangle			g_preview_zoom_buf = {0, 0, 0, 0};
static gboolean				g_preview_is_dragging = FALSE;
static FreehandEditModeType	g_preview_freehand_edit_mode_type = FREEHAND_EDIT_MODE_NONE;
static gboolean				g_preview_freehand_is_editing = FALSE;

static void init_preview_buf (PluginVals *vals);
static void preview_buf_clear ();

// scale of 100% zoom view.
// if image or selection size is smaller than or equal to preview size
// then the value will 1.0. 
// this value will used to original(global) point to cairo (user, local) point.
static gdouble		g_preview_buf_default_scale = 1.0;

gdouble
preview_global_to_local (gdouble pixel)
{
	return pixel * g_preview_buf_default_scale * (g_preview_zoom / 100.0);
}

gdouble
preview_local_to_global (gdouble pixel)
{
	return pixel / g_preview_buf_default_scale / (g_preview_zoom / 100.0);
}

static cairo_t*
preview_cairo_create (GdkDrawable *drawable)
{
	cairo_t *cr = gdk_cairo_create (drawable);

	//set default scale
	cairo_scale (cr, g_preview_buf_default_scale, g_preview_buf_default_scale);

	//set zoom
	cairo_scale (cr, g_preview_zoom / 100.0, g_preview_zoom / 100.0);

	//set translate (move image to center)
	if (g_preview_buf != NULL) {
		gdouble bufw = gdk_pixbuf_get_width (g_preview_buf);
		gdouble bufh = gdk_pixbuf_get_height (g_preview_buf);
		gdouble prew = preview_local_to_global (g_preview_width);
		gdouble preh = preview_local_to_global (g_preview_height);

		cairo_translate (cr,
			(prew - bufw) / 2.0,
			(preh - bufh) / 2.0);
	}
	return cr;
}

gboolean
preview_is_dragging ()
{
	return g_preview_is_dragging;
}

static void 
preview_draw_base_image (cairo_t *cr, PluginVals *vals)
{
	gdk_cairo_set_source_pixbuf (cr,
		g_preview_buf,
		0,
		0);
	cairo_paint (cr);
}

static void
preview_draw_line (cairo_t *cr, PluginVals *vals)
{
	cairo_set_source_rgba (cr, 
						   vals->line_color.r,
						   vals->line_color.g,
						   vals->line_color.b,
						   vals->line_color.a);
	switch (vals->area_type) {
		case AREA_TYPE_RADIATION: radiation_draw_line (cr, vals); break;
		case AREA_TYPE_PARALLEL:  parallel_draw_line (cr, vals); break;
	}
}

static void
preview_draw_grid (cairo_t *cr, PluginVals *vals)
{
	//TODO: is needed?
}


static void
preview_draw_guide (cairo_t *cr, PluginVals *vals)
{
	switch (vals->area_type) {
		case AREA_TYPE_RADIATION: radiation_draw_guide (cr, vals); break;
		case AREA_TYPE_PARALLEL:  parallel_draw_guide (cr, vals); break;
	}
}

static void
preview_realize_callback (GtkWidget  *widget,
						  PluginVals *vals)
{
	//do nothing
}

static gboolean
preview_event_callback (GtkWidget  *widget,
						GdkEvent   *event,
						PluginVals *vals)
{
	gboolean ret = FALSE;
	cairo_t *cr = preview_cairo_create (event->any.window);

	switch (event->type) {
		case GDK_BUTTON_PRESS:
			g_preview_is_dragging = TRUE;
			if (g_preview_freehand_edit_mode_type != FREEHAND_EDIT_MODE_NONE) {
				g_preview_freehand_is_editing = TRUE;
			}
			break;
		case GDK_BUTTON_RELEASE:
			g_preview_is_dragging = FALSE;
			g_preview_freehand_is_editing = FALSE;
			break;
		case GDK_SCROLL:
			if (g_preview_is_dragging == FALSE) {
				switch (event->scroll.direction) {
					case GDK_SCROLL_UP:
						preview_zoom_in ();
						return FALSE;
					case GDK_SCROLL_DOWN:
						preview_zoom_out ();
						return FALSE;
				}
			}
	}

	switch (vals->area_type) {
		case AREA_TYPE_RADIATION: ret = radiation_event_hook (cr, event, vals); break;
		case AREA_TYPE_PARALLEL:  ret =  parallel_event_hook (cr, event, vals); break;
	}
	cairo_destroy (cr);
	return ret;
}


static gboolean
preview_expose_callback (GtkWidget		*widget,
						 GdkEventExpose	*event,
						 PluginVals		*vals)
{
	cairo_t *cr = preview_cairo_create (event->window);

	// draw base image
	preview_draw_base_image (cr, vals);

	// draw lines
	preview_draw_line (cr, vals);

	// draw grid
	preview_draw_grid (cr, vals);

	// draw area guide
	preview_draw_guide (cr, vals);

	cairo_destroy (cr);
	return FALSE;
}

static void
preview_destroy_callback ()
{
	preview_buf_clear ();
}

GtkWidget*
preview_new (PluginVals	*vals)
{
	GdkColor gray = {0, 17767, 17767, 17767};

	g_preview = gtk_drawing_area_new ();

	gtk_widget_set_events (g_preview,	GDK_BUTTON_PRESS_MASK |
										GDK_BUTTON_RELEASE_MASK |
										GDK_BUTTON_MOTION_MASK |
										GDK_KEY_PRESS_MASK |
										GDK_KEY_RELEASE_MASK);
	
	g_signal_connect (G_OBJECT (g_preview), "realize",      G_CALLBACK (preview_realize_callback), vals);
	g_signal_connect (G_OBJECT (g_preview), "event",        G_CALLBACK (preview_event_callback),   vals);
	g_signal_connect (G_OBJECT (g_preview), "expose-event", G_CALLBACK (preview_expose_callback),  vals);
	g_signal_connect (G_OBJECT (g_preview), "destroy",      G_CALLBACK (preview_destroy_callback), vals);

	gtk_widget_set_size_request (g_preview, g_preview_width, g_preview_height);
	gtk_widget_modify_bg(g_preview, GTK_STATE_NORMAL, &gray);

	init_preview_buf (vals);

	return g_preview;
}

void
preview_zoom_in ()
{
	if (g_preview_zoom < 100) {
		g_preview_zoom += 25;
	}
	gtk_widget_set_sensitive (widgets.preview_plus_button, (g_preview_zoom < 100));
	gtk_widget_set_sensitive (widgets.preview_minus_button, TRUE);
	gtk_widget_queue_draw (widgets.preview);
	return;
}

void
preview_zoom_out ()
{
	if (g_preview_zoom > 25 ) {
		g_preview_zoom -= 25;
	}
	gtk_widget_set_sensitive (widgets.preview_plus_button, TRUE);
	gtk_widget_set_sensitive (widgets.preview_minus_button, (g_preview_zoom > 25));
	gtk_widget_queue_draw (widgets.preview);
	return;
}

gint
preview_zoom_get ()
{
	return g_preview_zoom;
}

FreehandEditModeType
preview_freehand_edit_mode_get ()
{
	return g_preview_freehand_edit_mode_type;
}

void
preview_freehand_edit_mode_set (FreehandEditModeType type)
{
	static GdkCursor *cursor_arrow = NULL;
	static GdkCursor *cursor_pencil = NULL;
	if (cursor_arrow == NULL) {
		cursor_arrow  = gdk_cursor_new (GDK_ARROW);
		cursor_pencil = gdk_cursor_new (GDK_PENCIL);
	}

	g_preview_freehand_edit_mode_type = type;

	// set cursor
	if (type == FREEHAND_EDIT_MODE_NONE) {
		gdk_window_set_cursor (gtk_widget_get_window (g_preview), cursor_arrow);
		g_preview_freehand_is_editing = FALSE;
	} else {
		gdk_window_set_cursor (gtk_widget_get_window (g_preview), cursor_pencil);
		g_preview_freehand_is_editing = FALSE;
	}
}

gboolean
preview_freehand_is_editing ()
{
	return g_preview_freehand_is_editing;
}


static void
preview_buf_clear ()
{
	if (g_preview_buf != NULL) {
		g_object_unref (g_preview_buf);
		g_preview_buf = NULL;
	}
	g_preview_zoom = 100;
}

static void
init_preview_buf (PluginVals *vals)
{
	gint img_width = gimp_image_width (vals->image_ID);
	gint img_height = gimp_image_height (vals->image_ID);
	gint x1, y1, x2, y2;
	gint buf_w, buf_h;

	if (gimp_selection_is_empty (vals->image_ID) == TRUE) {
		x1 = 0;
		y1 = 0;
		x2 = img_width;
		y2 = img_height;
		if (img_width > img_height) {
			if (g_preview_width > img_width) {
				g_preview_buf_default_scale = 1;
				buf_w = img_width;
				buf_h = img_height;
			} else {
				g_preview_buf_default_scale = (double)g_preview_width / img_width;
				buf_w = (int)(g_preview_width);
				buf_h = (int)(g_preview_buf_default_scale * img_height);
			}
		} else {
			if (g_preview_height > img_height) {
				g_preview_buf_default_scale = 1;
				buf_w = img_width;
				buf_h = img_height;
			} else {
				g_preview_buf_default_scale = (double)g_preview_height / img_height;
				buf_h = (int)(g_preview_height);
				buf_w = (int)(g_preview_buf_default_scale * img_width);
			}
		}
	} else {
		gboolean sel_non_empty;
		gimp_selection_bounds (vals->image_ID, &sel_non_empty, &x1, &y1, &x2, &y2);
		if (x2 - x1 > y2 - y1) {
			if (g_preview_width > x2 - x1) {
				g_preview_buf_default_scale = 1;
				buf_w = x2 - x1;
				buf_h = y2 - y1;
			} else {
				g_preview_buf_default_scale = (double)g_preview_width / (x2 - x1);
				buf_h = (int)(g_preview_buf_default_scale * (y2 - y1));
				buf_w = (int)(g_preview_width);
			}
		} else {
			if (g_preview_height > y2 - y1) {
				g_preview_buf_default_scale = 1;
				buf_w = x2 - x1;
				buf_h = y2 - y1;
			} else {
				g_preview_buf_default_scale = (double)g_preview_height / (y2 - y1);
				buf_w = (int)(g_preview_buf_default_scale * (x2 - x1));
				buf_h = (int)(g_preview_height);
			}
		}
	}

	//store original selection range in gimp
	g_preview_orig_buf.x = x1;
	g_preview_orig_buf.y = y1;
	g_preview_orig_buf.width  = x2 - x1;
	g_preview_orig_buf.height = y2 - y1;

	{
		gint32 templayer = gimp_layer_new_from_visible (vals->image_ID, vals->image_ID, PLUGIN_NAME "_PREVIEW");
		GdkPixbuf *temp;

		temp = gimp_drawable_get_sub_thumbnail (
			templayer,
			x1, 
			y1, 
			x2 - x1, 
			y2 - y1, 
			buf_w, 
			buf_h,
			GIMP_PIXBUF_LARGE_CHECKS);

		gimp_drawable_delete (templayer);

		if (buf_w < g_preview_orig_buf.width ||
			buf_h < g_preview_orig_buf.height) {
			g_preview_buf = gdk_pixbuf_scale_simple (
				temp,
				g_preview_orig_buf.width, 
				g_preview_orig_buf.height, 
				GDK_INTERP_NEAREST);
			g_object_unref (temp);
		} else {
			g_preview_buf = temp;
		}
	}
}

/* vim: set fdm=marker ts=4 sw=4 noet :*/
