/*
 * GIMP Speedline Plug-in -- This is a plug-in for The GIMP 2
 * Copyright (C) 2013-2015 OHTSUKA, Yoshio <ohtsuka.yoshio@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#include "speedline_preview.h"
#include "speedline_radiation.h"
#include "speedline_line.h"

typedef enum _RadiationGuideObj {
	RADIATION_GUIDE_OBJ_NONE = 0,
	RADIATION_GUIDE_OBJ_CENTER,
	RADIATION_GUIDE_OBJ_IN_CIRCUM,
	RADIATION_GUIDE_OBJ_OUT_CIRCUM,
	RADIATION_GUIDE_OBJ_OBLATENESS,
} RadiationGuideObj;

#define GET_DISTANCE(X1,Y1,X2,Y2) (sqrt(pow((X2)-(X1),2)+pow((Y2)-(Y1),2)))
#define GET_ELLIPSE_R(RAD,A,B) (sqrt(1/(pow(cos((RAD))/(A),2)+pow(sin(RAD)/(B),2))))
#define SWAP(TYPE,A,B) {TYPE temp = (B); (B) = (A); (A) = temp;}

static gdouble get_out_oblateness (gdouble in_radius, gdouble out_radius, gdouble oblateness);
static gdouble get_in_oblateness (gdouble in_radius, gdouble out_radius, gdouble oblateness);
static gdouble get_mouse_move_length (gdouble radius, gdouble oblateness, gdouble cx, gdouble cy, gdouble gx, gdouble gy);

static void path_guide_center (cairo_t *cr, gdouble cx, gdouble cy, gdouble in_radius, gdouble out_radius, gdouble oblateness);
static void path_guide_in_circum (cairo_t *cr, gdouble cx, gdouble cy, gdouble in_radius, gdouble out_radius, gdouble oblateness);
static void path_guide_out_circum (cairo_t *cr, gdouble cx, gdouble cy, gdouble in_radius, gdouble out_radius, gdouble oblateness);
static void path_guide_oblateness (cairo_t *cr, gdouble cx, gdouble cy, gdouble in_radius, gdouble out_radius, gdouble oblateness);
static void path_guide_in_freehand (cairo_t *cr, gdouble cx, gdouble cy, gdouble in_radius, gdouble out_radius, gdouble oblateness, KeyValue *points, gint *count);
static void path_guide_out_freehand (cairo_t *cr, gdouble cx, gdouble cy, gdouble in_radius, gdouble out_radius, gdouble oblateness, KeyValue *points, gint *count);
static RadiationGuideObj path_guide_hit_test (cairo_t *cr, GdkWindow *window, PluginVals *vals, gdouble x, gdouble y);
static void get_global_vals (PluginVals *vals, gdouble *cx, gdouble *cy, gdouble *in_radius, gdouble *out_radius, gdouble *oblateness);
static gdouble get_freehand_rad_value (gdouble rad, KeyValue *points, gint count);


static gdouble
get_out_oblateness (gdouble in_radius,
					gdouble out_radius,
					gdouble oblateness)
{
	return oblateness;
}

static gdouble
get_in_oblateness (gdouble in_radius,
				   gdouble out_radius,
				   gdouble oblateness)
{
	return 1.0 - in_radius/(out_radius/(1.0-oblateness) - (out_radius-in_radius));
}

static gdouble
get_mouse_move_length (gdouble radius,
					   gdouble oblateness,
					   gdouble cx,
					   gdouble cy,
					   gdouble gx,
					   gdouble gy)
{
	gdouble a = radius / (1 - oblateness);
	gdouble b = radius;
	gdouble n = (gy - cy) / (gx - cx);
	gdouble nx = sqrt((a*a*b*b) / (b*b + a*a*n*n)) * ((gx - cx) < 0 ? -1 : 1);
	gdouble ny = n * nx;
	gdouble mx = nx + cx;
	gdouble my = ny + cy;
	gdouble dist = GET_DISTANCE (gx, gy, mx, my);
	if (GET_DISTANCE (cx, cy, gx, gy) < GET_DISTANCE (cx, cy, mx, my)) {
		dist = -dist;
	}
	return dist;
}

static void
path_guide_center (cairo_t *cr,
				   gdouble cx,
				   gdouble cy,
				   gdouble in_radius,
				   gdouble out_radius,
				   gdouble oblateness)
{
	gdouble size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;
	cairo_arc (cr, cx, cy, preview_local_to_global (size), 0, G_PI * 2);
}

static void
path_guide_in_circum (cairo_t *cr,
					  gdouble cx,
					  gdouble cy,
					  gdouble in_radius,
					  gdouble out_radius,
					  gdouble oblateness)
{
	cairo_save (cr);
	cairo_translate (cr, cx, cy);
	cairo_scale (cr, 1 / (1 - get_in_oblateness (in_radius, out_radius, oblateness)), 1);
	cairo_arc (cr, 0, 0, in_radius, 0, G_PI * 2);
	cairo_restore (cr);
}

static void
path_guide_out_circum (cairo_t *cr,
					   gdouble cx,
					   gdouble cy,
					   gdouble in_radius,
					   gdouble out_radius,
					   gdouble oblateness)
{
	cairo_save (cr);
	cairo_translate (cr, cx, cy);
	cairo_scale (cr, 1 / (1 - get_out_oblateness (in_radius, out_radius, oblateness)), 1);
	cairo_arc (cr, 0, 0, out_radius, 0, G_PI * 2);
	cairo_restore (cr);
}

static void
path_guide_oblateness (cairo_t *cr,
					   gdouble cx,
					   gdouble cy,
					   gdouble in_radius,
					   gdouble out_radius,
					   gdouble oblateness)
{
	gdouble size;
	size = PREVIEW_GUIDE_POINT_SIZE/2;

	cairo_save (cr);
	cairo_translate (cr, cx, cy);
	cairo_arc (cr, out_radius * (1 / (1 - get_out_oblateness (in_radius, out_radius, oblateness))), 0, 
		preview_local_to_global (size), 0, G_PI * 2);
	cairo_restore (cr);
}

static void
path_guide_in_freehand (cairo_t		*cr,
						gdouble		cx,
						gdouble		cy,
						gdouble		in_radius,
						gdouble		out_radius,
						gdouble		oblateness,
						KeyValue	*points,
						gint		*count)
{
	gint i;
	gdouble rad1 = G_PI * 2 / FREEHAND_POINTS_MAX;
	gdouble oblat = get_in_oblateness (in_radius, out_radius, oblateness);
	gdouble a = in_radius / (1 - get_in_oblateness (in_radius, out_radius, oblateness));
	gdouble b = in_radius;
	gdouble stx = 0, sty = 0;

	if (*count <= 0) {
		radiation_reset_freehand (points, count);
	}

	cairo_save (cr);
	cairo_translate (cr, cx, cy);
	for (i = 0; i < *count; i++) {
		gdouble rad = points[i].key;
		gdouble rate = points[i].value;
		gdouble r = GET_ELLIPSE_R (rad, a, b);
		gdouble x = r * cos(rad) * rate;
		gdouble y = r * sin(rad) * rate;

		if (i == 0) {
			cairo_move_to (cr, x, y);
			stx = x;
			sty = y;
		} else {
			cairo_line_to (cr, x, y);
		}
	}
	cairo_line_to (cr, stx, sty);
	cairo_restore (cr);
}

static void
path_guide_out_freehand (cairo_t	*cr,
						 gdouble	cx,
						 gdouble	cy,
						 gdouble	in_radius,
						 gdouble	out_radius,
						 gdouble	oblateness,
						 KeyValue	*points,
						 gint		*count)
{
	gint i;
	gdouble rad1 = G_PI * 2 / FREEHAND_POINTS_MAX;
	gdouble oblat = get_out_oblateness (in_radius, out_radius, oblateness);
	gdouble a = out_radius / (1 - get_out_oblateness (in_radius, out_radius, oblateness));
	gdouble b = out_radius;
	gdouble stx = 0, sty = 0;

	if (*count <= 0) {
		radiation_reset_freehand (points, count);
	}

	cairo_save (cr);
	cairo_translate (cr, cx, cy);
	for (i = 0; i < *count; i++) {
		gdouble rad = points[i].key;
		gdouble rate = points[i].value;
		gdouble r = GET_ELLIPSE_R (rad, a, b);
		gdouble x = r * cos(rad) * rate;
		gdouble y = r * sin(rad) * rate;

		if (i == 0) {
			cairo_move_to (cr, x, y);
			stx = x;
			sty = y;
		} else {
			cairo_line_to (cr, x, y);
		}
	}
	cairo_line_to (cr, stx, sty);
	cairo_restore (cr);
}

static RadiationGuideObj
path_guide_hit_test (cairo_t	*cr,
					 GdkWindow	*window,
					 PluginVals	*vals,
					 gdouble	x,
					 gdouble	y)
{
	gdouble cx, cy, in_radius, out_radius, oblateness;

	get_global_vals (vals, &cx, &cy, &in_radius, &out_radius, &oblateness);

	cairo_set_source_rgba (cr, 1, 1, 1, 1);
	cairo_set_line_width (cr, preview_local_to_global (PREVIEW_GUIDE_LINE_WIDTH_FOR_HITTEST));
	cairo_device_to_user (cr, &x, &y);
	
	//check if out circum
	path_guide_oblateness (cr, cx, cy, in_radius, out_radius, oblateness);
	if (cairo_in_fill (cr, x, y)) {
		return RADIATION_GUIDE_OBJ_OBLATENESS;
	}
	cairo_new_path (cr);

	//check if out circum
	path_guide_out_circum (cr, cx, cy, in_radius, out_radius, oblateness);
	if (cairo_in_stroke (cr, x, y)) {
		return RADIATION_GUIDE_OBJ_OUT_CIRCUM;
	}
	cairo_new_path (cr);

	//check if in circum
	path_guide_in_circum (cr, cx, cy, in_radius, out_radius, oblateness);
	if (cairo_in_stroke (cr, x, y)) {
		return RADIATION_GUIDE_OBJ_IN_CIRCUM;
	}
	cairo_new_path (cr);

	//check if center
	path_guide_center (cr, cx, cy, in_radius, out_radius, oblateness);
	if (cairo_in_fill (cr, x, y)) {
		return RADIATION_GUIDE_OBJ_CENTER;
	}
	cairo_new_path (cr);

	return RADIATION_GUIDE_OBJ_NONE;
}

static void
get_global_vals (PluginVals	*vals,
				 gdouble	*cx,
				 gdouble	*cy,
				 gdouble	*in_radius,
				 gdouble	*out_radius,
				 gdouble	*oblateness)
{
	gdouble resx, resy;
	gimp_image_get_resolution (vals->image_ID, &resx, &resy);
	*cx = gimp_units_to_pixels (vals->radiation_x, vals->area_unit, resx);
	*cy = gimp_units_to_pixels (vals->radiation_y, vals->area_unit, resy);
	*in_radius  = gimp_units_to_pixels (vals->radiation_in_radius, vals->area_unit, resx);
	*out_radius = gimp_units_to_pixels (vals->radiation_out_radius, vals->area_unit, resx);
	*oblateness = vals->radiation_oblateness;
}

static void
update_freehand_points (PluginVals				*vals,
						FreehandEditModeType	type,
						gdouble					stx,
						gdouble					sty,
						gdouble					edx,
						gdouble					edy)
{
	gint i;
	gdouble in_radius = vals->radiation_in_radius;
	gdouble out_radius = vals->radiation_out_radius;
	gdouble oblateness = vals->radiation_oblateness;
	KeyValue *points;
	gint points_count;
	gdouble a, b;
	gdouble rad1 = G_PI * 2 / FREEHAND_POINTS_MAX;
	gdouble rad_st, rad_ed;
	gdouble dist_st, dist_ed;

	switch (type) {
		case FREEHAND_EDIT_MODE_RADIATION_IN:
			points = vals->radiation_in_freehand_points;
			points_count = vals->radiation_in_freehand_points_count;
			a = in_radius / (1 - get_in_oblateness (in_radius, out_radius, oblateness));
			b = in_radius;
			break;

		case FREEHAND_EDIT_MODE_RADIATION_OUT:
			points = vals->radiation_out_freehand_points;
			points_count = vals->radiation_out_freehand_points_count;
			a = out_radius / (1 - get_out_oblateness (in_radius, out_radius, oblateness));
			b = out_radius;
			break;

		default:
			//do nothing
			return;
	}

	stx -= vals->radiation_x;
	sty -= vals->radiation_y;
	edx -= vals->radiation_x;
	edy -= vals->radiation_y;

	rad_st = atan2 (sty, stx);
	rad_ed = atan2 (edy, edx);

	if (rad_st < 0) { rad_st += G_PI * 2; }
	if (rad_ed < 0) { rad_ed += G_PI * 2; }

	dist_st = GET_DISTANCE (0, 0, stx, sty);
	dist_ed = GET_DISTANCE (0, 0, edx, edy);

	if (rad_ed < rad_st) {
		SWAP(gdouble, rad_st, rad_ed);
		SWAP(gdouble, dist_st, dist_ed);
	}

	// check if step over rad=0
	if (rad_st + G_PI * 2 - rad_ed < rad_ed - rad_st) {
		if (G_PI * 2 - rad_ed + rad_st > rad1) {
			gdouble n = (dist_ed - dist_st) / ((G_PI * 2 - rad_ed + rad_st) / rad1);
			gdouble nn = 0;

			for (i = 0; i < points_count; i++) {
				gdouble rad = points[i].key;
				gdouble rate = points[i].value;
				if (rad < rad_ed) continue;
				points[i].value = (dist_ed + nn) / GET_ELLIPSE_R (rad, a, b);
				nn += n;
			}

			for (i = 0; i < points_count; i++) {
				gdouble rad = points[i].key;
				gdouble rate = points[i].value;
				points[i].value = (dist_ed + nn) / GET_ELLIPSE_R (rad, a, b);
				nn += n;
				if (rad >= rad_st) break;
			}
		} else {
			for (i = 0; i < points_count; i++) {
				gdouble rad = points[i].key;
				gdouble rate = points[i].value;
				if (rad >= rad_st) {
					points[i].value = dist_st / GET_ELLIPSE_R (rad, a, b);
					break;
				}
			}
		}
	} else {
		if (rad_ed - rad_st > rad1) {
			gdouble n = (dist_ed - dist_st) / ((rad_ed - rad_st) / rad1);
			gdouble nn = 0;

			for (i = 0; i < points_count; i++) {
				gdouble rad = points[i].key;
				gdouble rate = points[i].value;
				if (rad >= rad_st && rad <= rad_ed) {
					points[i].value = (dist_st + nn) / GET_ELLIPSE_R (rad, a, b);
					nn += n;
				}
			}
		} else {
			for (i = 0; i < points_count; i++) {
				gdouble rad = points[i].key;
				gdouble rate = points[i].value;
				if (rad >= rad_st) {
					points[i].value = dist_st / GET_ELLIPSE_R (rad, a, b);
					break;
				}
			}
		}
	}
}

static gdouble
get_freehand_rad_value (gdouble		rad,
						KeyValue	*points,
						gint		count)
{
	gint i;
	gdouble rad1 = G_PI * 2 / count;

	if (rad < 0) {
		rad += G_PI * 2;
	}

	if (rad > G_PI * 2 - rad1) {
		return points[0].value;
	}

	for (i = 0; i < count; i++) {
		if (points[i].key >= rad) {
			return points[i].value;
		}
	}
	return 0;
}

void
radiation_draw_guide (cairo_t		*cr,
					  PluginVals	*vals)
{
	gint n, nmax;
	gdouble sx, sy, dx, dy;
	gdouble cx, cy, in_radius, out_radius, oblateness;

	get_global_vals (vals, &cx, &cy, &in_radius, &out_radius, &oblateness);

	cairo_set_line_width (cr, preview_local_to_global (PREVIEW_GUIDE_LINE_WIDTH));

	cairo_set_source_rgba (cr,
		PREVIEW_GUIDE_COLOR_R,
		PREVIEW_GUIDE_COLOR_G,
		PREVIEW_GUIDE_COLOR_B,
		1);

	// in freehand
	if (vals->radiation_in_freehand_active) {
		path_guide_in_freehand (cr, cx, cy, in_radius, out_radius, oblateness, 
								vals->radiation_in_freehand_points,
								&vals->radiation_in_freehand_points_count);
		cairo_save (cr);
		{
			gdouble dashes[] = {6, 3};
			if (preview_freehand_edit_mode_get () == FREEHAND_EDIT_MODE_RADIATION_IN) {
				cairo_set_source_rgba (cr,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_R,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_G,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_B,
					1);
			}
			cairo_set_dash (cr, dashes, 2, 0.0);
			cairo_stroke (cr);
		}
		cairo_restore (cr);
	}

	// out freehand
	if (vals->radiation_out_freehand_active) {
		path_guide_out_freehand (cr, cx, cy, in_radius, out_radius, oblateness,
								 vals->radiation_out_freehand_points,
								 &vals->radiation_out_freehand_points_count);
		cairo_save (cr);
		{
			gdouble dashes[] = {6, 3};
			if (preview_freehand_edit_mode_get () == FREEHAND_EDIT_MODE_RADIATION_OUT) {
				cairo_set_source_rgba (cr,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_R,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_G,
					PREVIEW_GUIDE_FREEHAND_EDIT_COLOR_B,
					1);
			}
			cairo_set_dash (cr, dashes, 2, 0.0);
			cairo_stroke (cr);
		}
		cairo_restore (cr);
	}

	//oblateness
	path_guide_oblateness (cr, cx, cy, in_radius, out_radius, oblateness);
	cairo_fill (cr);

	//out
	path_guide_out_circum (cr, cx, cy, in_radius, out_radius, oblateness);
	cairo_stroke (cr);

	//in
	path_guide_in_circum (cr, cx, cy, in_radius, out_radius, oblateness);
	cairo_stroke (cr);

	//center
	path_guide_center (cr, cx, cy, in_radius, out_radius, oblateness);
	cairo_fill (cr);

	//iines
	sx = cx;
	sy = cy;
	dx = cx + out_radius;
	dy = cy;
	nmax = PREVIEW_GUIDE_LINE_NUMBER;
	cairo_save (cr);
	cairo_translate (cr, cx, cy);
	for (n = 0; n < nmax; n++) {
		// X : CenterX + R * Math.cos(Radian)
		// Y : CenterY - R * Math.sin(Radian)
		sx = out_radius * cos ((gdouble)n/nmax * G_PI * 2);
		sy = out_radius * sin ((gdouble)n/nmax * G_PI * 2);
		dx = in_radius * cos ((gdouble)n/nmax * G_PI * 2);
		dy = in_radius * sin ((gdouble)n/nmax * G_PI * 2);
		cairo_save (cr);
		cairo_scale (cr, 1 / (1 - get_out_oblateness (in_radius, out_radius, oblateness)), 1);
		cairo_move_to (cr, sx, sy);
		cairo_restore (cr);

		cairo_save (cr);
		cairo_scale (cr, 1 / (1 - get_in_oblateness (in_radius, out_radius, oblateness)), 1);
		cairo_line_to (cr, dx, dy);
		cairo_restore (cr);
	}
	cairo_restore (cr);
	cairo_stroke (cr);
}

void
radiation_draw_line (cairo_t	*cr,
					 PluginVals	*vals)
{
	LineObj *line = line_new (vals);
	gdouble cx, cy, in_radius, out_radius, oblateness;
	gdouble oa, ob, ia, ib;
	gdouble block_sum = 0;
	gdouble wave_len = 0;
	gdouble* wave_div_rad = NULL;
	gint wave_div_i = 0;
	gdouble wave_div_pre = 0;

	get_global_vals (vals, &cx, &cy, &in_radius, &out_radius, &oblateness);

	oa = out_radius * 1 / (1 - get_out_oblateness (in_radius, out_radius, oblateness));
	ob = out_radius;
	ia = in_radius * 1 / (1 - get_in_oblateness (in_radius, out_radius, oblateness));
	ib = in_radius;

	cairo_save (cr);
	{
		gdouble stx, sty, edx, edy, rad;
		gdouble psum = 0;

		//normalize center point
		cairo_translate (cr, cx, cy);

		//get start point and some parameter
		switch (vals->line_pitch_type) {
			case LINE_PITCH_TYPE_IN: 
				stx = ia;
				sty = 0;
				rad = 0;
				break;
			case LINE_PITCH_TYPE_OUT:
				stx = oa;
				sty = 0;
				rad = 0;
				break;
		}

		if (line_is_wave_mode (line) == TRUE) {
			int i;
			int npoint = line_wave_npoint_get (line);

			wave_div_rad = g_malloc0_n(sizeof(gdouble), npoint);

			if (oblateness == 0) {
				// circle
				for (i = 0; i < npoint; i++) {
					wave_div_rad[i] = (G_PI * 2 / npoint) * (i + 1);
				}
			} else {
				// ellipse
				gint segnum = 256;
				gdouble rad_total = 0;
				gdouble dist_total = 0;
				gdouble dist1;
				gdouble rad1 = (G_PI_2) / segnum; 
				gdouble x, y, pre_x, pre_y;
				gdouble a = oa;
				gdouble b = ob;
				gdouble* dist_list = g_malloc0_n (sizeof(gdouble), segnum);
				gdouble* rad_list = g_malloc0_n (sizeof(gdouble), segnum);
				gint n, n4;

				n4 = npoint / 4;
				
				// segnum polygon
				rad_total = rad1;
				pre_x = a * cos(0);
				pre_y = b * sin(0);
				for (i = 0; i < segnum; i++) {
					x = a * cos(rad_total);
					y = b * sin(rad_total);
					dist_list[i] = GET_DISTANCE (x, y, pre_x, pre_y);
					dist_total += dist_list[i];
					rad_total += rad1;
					rad_list[i] = rad_total;
					pre_x = x;
					pre_y = y;
				}

				// length of one arc
				dist1 = dist_total / n4;

				// divide elliple(1/4) by dist1 length
				n = 0;
				dist_total = 0;
				for (i = 0; i < segnum; i++) {
					dist_total += dist_list[i];
					if (dist_total > dist1) {
						wave_div_rad[n] = rad1 * i;
						n++;
						dist_total = 0;
						if (n > n4) break;
					}
				}
				wave_div_rad[n] = G_PI_2;

				// copy 1/4 data to 2/4, 3/4, 4/4
				for (i = 0; i < n4; i++) {
					wave_div_rad[i + n4 * 2] = wave_div_rad[i] + G_PI;
				}
				for (i = 0; i < n4; i++) {
					if ((n4*2 -i -1) % n4 == 0) {
						wave_div_rad[n4*2 -1] = G_PI;
						wave_div_rad[n4*4 -1] = G_PI * 2;
					} else {
						wave_div_rad[n4*2 -i -2] = G_PI     - wave_div_rad[i];
						wave_div_rad[n4*4 -i -2] = G_PI * 2 - wave_div_rad[i];
					}
				}

				g_free(dist_list);
				g_free(rad_list);
			}
		}

		// foreach lines
		while (rad < G_PI * 2) {
			gdouble x1, y1, x2, y2;

			//get line point (x1,y1) (x2,y2)
			{
				switch (vals->line_pitch_type) {
					case LINE_PITCH_TYPE_IN:
						x1 = stx;
						y1 = sty;
						x2 = oa * cos(rad);
						y2 = ob * sin(rad);
						break;
					case LINE_PITCH_TYPE_OUT:
						x1 = ia * cos(rad);
						y1 = ib * sin(rad);
						x2 = stx;
						y2 = sty;
						break;
				}
			}

			if (line_is_block_mode (line) == TRUE) {
				if (psum >= block_sum) {
					line_block_state_invert(line);
					if (line_block_state_get (line) == TRUE) {
						block_sum += line_block_on_pitch_get (line);
					} else {
						block_sum += line_block_off_pitch_get (line);
					}
				}
				if (line_block_state_get (line) == FALSE) {
					goto block_skip;
				}
			}

			//draw line
			cairo_save (cr);
			{
				gdouble line_length;
				cairo_translate (cr, x1, y1);
				cairo_rotate (cr, atan2((y2 - y1), (x2 - x1)));
				line_length = GET_DISTANCE (x1 ,y1 ,x2 ,y2);

				// Wave mode
				if (line_is_wave_mode (line)) {
					gdouble r, depth;
					if (rad > wave_div_rad[wave_div_i]) {
						wave_div_i++;
					}
					if (wave_div_i == 0) {
						r = fmod(rad, wave_div_rad[wave_div_i]) / wave_div_rad[wave_div_i];
					} else {
						gdouble pre = wave_div_rad[wave_div_i - 1];
						r = fmod(rad - pre, wave_div_rad[wave_div_i] - pre) / (wave_div_rad[wave_div_i] - pre);
					}
					depth = line_wave_depth_get (line, r);
					cairo_translate (cr, depth, 0);
				}
				
				// Freehand mode In
				if (vals->radiation_in_freehand_active) {
					gdouble rate = get_freehand_rad_value (atan2(y1, x1),
														   vals->radiation_in_freehand_points,
														   vals->radiation_in_freehand_points_count);
					gdouble dist_in = GET_DISTANCE (0, 0, x1, y1);
					cairo_translate (cr, dist_in * rate - dist_in, 0);
					line_length -= dist_in * rate - dist_in;
				}

				// Freehand mode Out
				if (vals->radiation_out_freehand_active) {
					gdouble rate = get_freehand_rad_value (atan2(y2, x2),
														   vals->radiation_out_freehand_points,
														   vals->radiation_out_freehand_points_count);
					gdouble dist_out = GET_DISTANCE (0, 0, x2, y2);
					line_length += dist_out * rate - dist_out;
				}

				line_draw (line, cr, line_length);
			}
			cairo_restore (cr);

block_skip:

			//compute next point
			{
				gdouble a, b;
				gdouble rad_st, rad_incr;
				gdouble pitch;
				gint hit = 0;

				switch (vals->line_pitch_type) {
					case LINE_PITCH_TYPE_IN:
						a = ia;
						b = ib;
						break;
					case LINE_PITCH_TYPE_OUT:
						a = oa;
						b = ob;
						break;
				}

				rad_incr = 10 * G_PI/180;
				rad_st = rad;
				do {
					pitch = line->pitch - line->pitch * g_rand_double (line->randobj_pitch) * line->pitch_rand / 100.0;
				} while (pitch <= 0);
				if (pitch < 0.01) {
					pitch = 0.01;
				}

				while (hit < 8) {
					gdouble dist;

					rad += rad_incr;

					edx = a * cos(rad);
					edy = b * sin(rad);
					dist = GET_DISTANCE (stx, sty, edx, edy);
					if (rad_incr == 0) {
						rad_incr = G_PI/180;
						rad += rad_incr;
						continue;
					}
					if (dist < pitch && rad_incr > 0) {
						//not hit
					} else if (dist < pitch && rad_incr < 0) {
						rad_incr = rad_incr / -2;
						hit++;
					} else if (dist > pitch && rad_incr > 0) {
						rad_incr = rad_incr / -2;
						hit++;
					} else if (dist > pitch && rad_incr < 0) {
						//not hit
					}
					if (rad >= G_PI * 3) {
						rad_incr = rad_incr / -2;
					}
				}
			}

			psum += GET_DISTANCE (stx, sty, edx, edy);

			//next lines parameter
			stx = edx;
			sty = edy;
			rad = rad;
		}
	}
	cairo_restore (cr);

	if (line_is_wave_mode (line) == TRUE) {
		if (wave_div_rad) g_free(wave_div_rad);
		wave_div_rad = NULL;
		wave_div_i = 0;
	}
	line_destroy (line);
}


gboolean
radiation_event_hook (cairo_t		*cr,
					  GdkEvent		*event,
					  PluginVals	*vals)
{
	static RadiationGuideObj dragging_obj = RADIATION_GUIDE_OBJ_NONE;
	static gdouble dragging_start_point_x = 0;
	static gdouble dragging_start_point_y = 0;
	static gdouble dragging_start_in_radius = 0;
	static gdouble dragging_start_out_radius = 0;
	static gdouble dragging_start_in_oblateness = 0;
	static gdouble dragging_start_out_oblateness = 0;
	static gdouble dragging_start_freehand_point_x = 0;
	static gdouble dragging_start_freehand_point_y = 0;

	switch (event->type) {
		case GDK_EXPOSE:
			break;

		case GDK_BUTTON_PRESS:
			dragging_start_point_x = event->button.x;
			dragging_start_point_y = event->button.y;
			dragging_start_in_radius = vals->radiation_in_radius;
			dragging_start_out_radius = vals->radiation_out_radius;
			dragging_start_in_oblateness = get_in_oblateness(vals->radiation_in_radius, 
															 vals->radiation_out_radius,
															 vals->radiation_oblateness);
			dragging_start_out_oblateness = vals->radiation_oblateness;
			cairo_device_to_user (cr, &dragging_start_point_x, &dragging_start_point_y);

			if (preview_freehand_is_editing ()) {
				dragging_start_freehand_point_x = dragging_start_point_x;
				dragging_start_freehand_point_y = dragging_start_point_y;
			} else {
				dragging_obj = path_guide_hit_test (cr,
													event->any.window, 
													vals,
													event->button.x,
													event->button.y);
			}
			break;

		case GDK_BUTTON_RELEASE:
			dragging_start_point_x = 0;
			dragging_start_point_y = 0;
			dragging_start_in_radius = 0;
			dragging_start_out_radius = 0;
			dragging_start_in_oblateness = 0;
			dragging_start_out_oblateness = 0;
			dragging_obj = RADIATION_GUIDE_OBJ_NONE;
			dragging_start_freehand_point_x = 0;
			dragging_start_freehand_point_y = 0;
			break;

		case GDK_MOTION_NOTIFY:
			{
				gdouble gx, gy;
				gdouble cx, cy;
				gdouble resx, resy;
				gdouble temp;
				gimp_image_get_resolution (vals->image_ID, &resx, &resy);
				cx = gimp_units_to_pixels (vals->radiation_x, vals->area_unit, resx);
				cy = gimp_units_to_pixels (vals->radiation_y, vals->area_unit, resy);

				gx = event->motion.x;
				gy = event->motion.y;
				cairo_device_to_user (cr, &gx, &gy);

				if (preview_freehand_is_editing ()) {
					update_freehand_points (vals,
											preview_freehand_edit_mode_get (),
											dragging_start_freehand_point_x,
											dragging_start_freehand_point_y,
											gx, gy);
					dragging_start_freehand_point_x = gx;
					dragging_start_freehand_point_y = gy;
				} else {
					switch (dragging_obj) {
						case RADIATION_GUIDE_OBJ_CENTER:
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_x), gimp_pixels_to_units (gx, vals->area_unit, resx));
							gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_y), gimp_pixels_to_units (gy, vals->area_unit, resy));
							break;

						case RADIATION_GUIDE_OBJ_IN_CIRCUM:
							{
								gdouble dist = get_mouse_move_length (dragging_start_in_radius, 
																	 dragging_start_in_oblateness, 
																	 cx, cy, gx, gy);
								temp = gimp_pixels_to_units (dragging_start_in_radius + dist, vals->area_unit, resx);
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_in_radius), temp);
							}
							break;

						case RADIATION_GUIDE_OBJ_OUT_CIRCUM:
							{
								gdouble dist = get_mouse_move_length (dragging_start_out_radius, 
																	 dragging_start_out_oblateness, 
																	 cx, cy, gx, gy);
								temp = gimp_pixels_to_units (dragging_start_out_radius + dist, vals->area_unit, resx);
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_out_radius), temp);

								// keep oblateness of in_radius
								{
									gdouble len = dragging_start_out_radius - dragging_start_in_radius + dist;
									gdouble a = dragging_start_in_radius / (1 - dragging_start_in_oblateness) + len;
									gdouble b = dragging_start_in_radius + len;
									gdouble ob = 1 - b / a;
									gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_oblateness), ob);
								}
							}
							break;

						case RADIATION_GUIDE_OBJ_OBLATENESS:
							{
								gdouble lr = gx - vals->radiation_x;
								gdouble out_oblateness, in_oblateness;

								out_oblateness = 1 - (vals->radiation_out_radius / lr);
								in_oblateness = get_in_oblateness (vals->radiation_in_radius, vals->radiation_out_radius, out_oblateness);
								gtk_spin_button_set_value (GTK_SPIN_BUTTON (widgets.radiation_oblateness), out_oblateness);
							}
							break;
					}
				}
				gtk_widget_queue_draw (widgets.preview);
			}
			break;

		case GDK_KEY_PRESS:
			break;

		case GDK_KEY_RELEASE:
			break;

		default:
			break;
	}

	return FALSE;
}

void
radiation_reset_freehand (KeyValue	*points,
						  gint		*count)
{
	gdouble rad1 = G_PI * 2 / FREEHAND_POINTS_MAX;
	gint i = 0;
	*count = FREEHAND_POINTS_MAX;

	for (i = 0; i < FREEHAND_POINTS_MAX; i++) {
		points[i].key = i * rad1;
		points[i].value = 1;
	}
}


/* vim: set fdm=marker ts=4 sw=4 noet :*/
